# Copyright (C) 2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Any
from typing import Callable
from typing import Coroutine
from typing import Generic
from typing import TypeVar

import bittensor
import fastapi
import fastapi.routing
import starlette
import starlette.requests

from ._request import SynapseRequest


S = TypeVar('S', bound=bittensor.Synapse)


class SynapseRouteHandler(fastapi.routing.APIRoute, Generic[S]):
    logger: logging.Logger = logging.getLogger(__name__)
    synapse_class: type[S]

    @classmethod
    def build(
        cls,
        synapse_class: type[Any]
    ) -> type['SynapseRouteHandler[Any]']:
        return type(f'{synapse_class.__name__}', (cls, ), {
            'synapse_class': synapse_class
        })

    def get_route_handler(self) -> Callable[[starlette.requests.Request], Coroutine[Any, Any, fastapi.Response]]:
        handler = super().get_route_handler()

        async def f(request: starlette.requests.Request) -> fastapi.Response:
            request = SynapseRequest(
                model=self.synapse_class,
                scope=request.scope,
                receive=request.receive
            )
            return await handler(request)
        return f


# Copyright (C) 2023-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
from typing import Any
from typing import TypeVar

from substrateinterface.utils.ss58 import ss58_decode
from substrateinterface.utils.ss58 import is_valid_ss58_address
from substrateinterface import Keypair
from pydantic_core import CoreSchema
from pydantic_core import core_schema
from pydantic.json_schema import JsonSchemaValue
from pydantic import GetJsonSchemaHandler


__all__: list[str] = [
    'SS58Address'
]

T = TypeVar('T', bound='SS58Address')


class SS58Address(str):
    __module__: str = 'canonical.ext.bittensor'

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=core_schema.str_schema(),
            python_schema=core_schema.chain_schema([
                core_schema.is_instance_schema(str),
                core_schema.no_info_plain_validator_function(cls.validate)
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(cls.serialize)
        )

    @functools.cached_property
    def public_bytes(self):
        return bytes.fromhex(ss58_decode(self))

    @functools.cached_property
    def keypair(self):
        return Keypair(ss58_address=self)

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.str_schema())

    @classmethod
    def validate(cls: type[T], instance: T | str) -> T:
        if not is_valid_ss58_address(instance):
            raise ValueError("not a valid SS58 address")
        return cls(instance)

    def serialize(self) -> str:
        return self

    def verify(
        self,
        data: str | bytes,
        signature: str | bytes
    ):
        return self.keypair.verify(data, signature)

    def __repr__(self):
        return f'<SS58Address: {str(self)}>'


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import inspect
import logging
import threading
from contextlib import AsyncExitStack
from typing import overload
from typing import Any
from typing import AsyncGenerator
from typing import Awaitable
from typing import Callable
from typing import ParamSpec
from typing import TypeAlias
from typing import TypeVar
from typing import TYPE_CHECKING

import bittensor
import fastapi
import fastapi.params
from fastapi.dependencies.utils import get_dependant
from fastapi.dependencies.utils import solve_dependencies

from canonical.ext.bittensor.axon import SynapseResponse
from canonical.ext.bittensor.models import SynapseEnvelope
from ._task import Task
if TYPE_CHECKING:
    from ._soma import SynapseHandlerType
    from ._soma import Soma


P = ParamSpec('P')
R = TypeVar('R')
S = TypeVar('S', bound=bittensor.Synapse)
SynapseHandlerMapping: TypeAlias = dict[str, 'SynapseHandlerType']


class TaskRunner:
    handlers: SynapseHandlerMapping
    logger: logging.Logger = logging.getLogger(__name__)
    running: set[asyncio.Task[None]]
    tasks: dict[str, Task[Any]]

    @overload
    async def run_handler(
        self,
        task: Task[Any],
        handler: Callable[P, AsyncGenerator[None | SynapseResponse, bittensor.Synapse | None]]
    ) -> AsyncGenerator[None | SynapseResponse, bittensor.Synapse | None]:
        ...

    @overload
    async def run_handler(
        self,
        task: Task[Any],
        handler: Callable[P, Awaitable[None | SynapseResponse]],
        envelope: SynapseEnvelope[Any]
    ) -> None | SynapseResponse:
        ...

    @property
    def pending(self) -> list[Task[Any]]:
        return [t for t in self.tasks.values() if t.is_runnable()]

    def __init__(
        self,
        soma: 'Soma[Any]',
        handlers: SynapseHandlerMapping,
        loop: asyncio.AbstractEventLoop
    ):
        self.handlers = handlers
        self.lock = threading.Lock()
        self.loop = loop
        self.running = set()
        self.soma = soma
        self.tasks = {}

    def get_handler(self, synapse: bittensor.Synapse) -> 'SynapseHandlerType':
        return self.handlers[type(synapse).__name__]

    def schedule(
        self,
        envelope: SynapseEnvelope[S],
        future: asyncio.Future[SynapseResponse]
    ) -> None:
        k = self.soma.generate_task_id(envelope.synapse)
        with self.lock:
            task: Task[Any] | None = self.tasks.get(k)
            if task is None:
                self.logger.debug(
                    "Received new task (id: %s, kind: %s)",
                    k[:6],
                    type(envelope.synapse).__name__
                )
                task = Task(
                    task_id=k,
                    runner=self,
                    handler=self.get_handler(envelope.synapse)
                )
                self.tasks[k] = task
            else:
                self.logger.debug(
                    "Subscribing to existing task (id: %s, envelope: %s)",
                    k[:6],
                    type(envelope.synapse).__name__
                )
            task.subscribe(envelope, future)

    async def run_all(self) -> None:
        pending = self.pending

        self.logger.debug("Running %s pending tasks", len(pending))
        for task in pending:
            self.logger.debug(
                "Scheduled task (id: %s)",
                task.task_id[:6]
            )
            self.running.add(self.loop.create_task(task.run()))
        if not self.running:
            return
        self.logger.debug("Running tasks: %s", len(self.running))
        done, self.running = await asyncio.wait(self.running, timeout=0.1)
        self.logger.debug("Completed %s tasks", len(done))

    async def run_handler(
        self,
        task: Task[Any],
        handler: Callable[P, R],
        envelope: SynapseEnvelope[Any] | None = None,
        request: fastapi.Request | None = None
    ) -> Any:
        async with AsyncExitStack() as stack:
            cache:  dict[tuple[Callable[..., Any], tuple[str]], Any] = {}
            dependant, kwargs = await self.inject_dependencies(
                stack=stack,
                task=task,
                handler=handler,
                envelope=envelope,
                request=request,
                cache=cache
            )
            assert dependant.call is not None
            assert callable(dependant.call)
            result = dependant.call(**kwargs)
            if inspect.isawaitable(result):
                result = await result
            return result

    async def inject_dependencies(
        self,
        stack: AsyncExitStack,
        task: Task[Any],
        handler: Callable[..., Any],
        envelope: SynapseEnvelope[S] | None,
        cache: dict[tuple[Callable[..., Any], tuple[str]], Any],
        request: fastapi.Request | None = None
    ):
        body = None
        path = '/'
        if envelope is not None:
            body = envelope.synapse.model_dump(mode='json')
            path = f'/{type(envelope.synapse).__name__}'
        request = self.request_factory(request, path, envelope)

        # Add the task to the request state so that it
        # can be injected later.
        request.state.task = task

        dependant = get_dependant(call=handler, path=path)
        values, errors, *_ = await solve_dependencies(
            request=request,
            dependant=dependant,
            body=body,
            dependency_overrides_provider=None,
            dependency_cache=cache,
            async_exit_stack=stack
        )
        if errors:
            raise ValueError(errors)
        cache.update(cache)
        assert dependant.call is not None
        assert callable(dependant.call)
        return dependant, values

    def request_factory(
        self,
        request: fastapi.Request | None,
        path: str,
        envelope: SynapseEnvelope[Any] | None = None
    ) -> fastapi.Request:
        if request is None:
            scope: dict[str, Any] = {
                'type': 'http',
                'path': path,
                'query_string': '',
                'headers': []
            }
            if envelope is not None:
                scope['client'] = (envelope.remote_host, envelope.remote_port)
            request = fastapi.Request(scope=scope)
        return request


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
from cryptography.hazmat.primitives.asymmetric import ed25519
from libcanonical.types import HexEncoded

from ._hotkeypublickey import HotkeyPublicKey


class HotkeyPrivateKey(HotkeyPublicKey):
    private_key: HexEncoded = pydantic.Field(
        default=...,
        alias='privateKey'
    )

    @property
    def private(self):
        return ed25519.Ed25519PrivateKey.from_private_bytes(self.private_key)


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic


class HotkeyRef(pydantic.BaseModel):
    model_config = {'populate_by_name': True}

    name: str = pydantic.Field(
        default=...
    )

    hotkey: str = pydantic.Field(
        default=...
    )

    @property
    def qualname(self):
        return f'{self.name}/{self.hotkey}'

    def __hash__(self):
        return hash(f'{self.name}/{self.hotkey}')


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import ipaddress
import re
import pathlib
import sys
import warnings
from typing import Any

import pydantic
from bittensor.core.settings import FINNEY_ENTRYPOINT
from bittensor.core.settings import FINNEY_TEST_ENTRYPOINT
from libcanonical.bases import EnvironmentBaseModel
from libcanonical.types import ColonSeparatedList
from libcanonical.types import HTTPResourceLocator
from libcanonical.types import WebSocketResourceLocator

from .models import ColdkeyRef
from .models import Hotkey


class BittensorEnvironment(EnvironmentBaseModel):
    axon_ip: ipaddress.IPv4Address = pydantic.Field(
        default=ipaddress.IPv4Address('0.0.0.0'),
        alias='BT_AXON_HOST'
    )

    axon_port: int = pydantic.Field(
        default=8091,
        alias='BT_AXON_PORT',
        gt=1000
    )

    axon_disabled: bool = pydantic.Field(
        default=False,
        alias='BT_AXON_DISABLED'
    )

    chain_endpoint: HTTPResourceLocator | WebSocketResourceLocator | None = pydantic.Field(
        default=None,
        alias='BT_CHAIN_ENDPOINT'
    )

    enabled_hotkeys: ColonSeparatedList[Hotkey] = pydantic.Field(
        default_factory=ColonSeparatedList,
        alias='BT_ENABLED_HOTKEYS'
    )

    network: str = pydantic.Field(
        default='finney',
        alias='BT_NETWORK'
    )

    netuid: int | None = pydantic.Field(
        default=None,
        alias='BT_SUBNET_UID'
    )

    force_validator_permit: bool = pydantic.Field(
        default=True,
        alias='BT_FORCE_VALIDATOR_PERMIT'
    )

    validator_min_stake: int = pydantic.Field(
        default=4096,
        alias='BT_VALIDATOR_MIN_STAKE'
    )

    wallet_path: pathlib.Path = pydantic.Field(
        default=pathlib.Path('~/.bittensor/wallets').expanduser(),
        alias='BT_WALLET_PATH'
    )

    @pydantic.field_validator('enabled_hotkeys', mode='after')
    def preprocess_enabled_hotkeys(cls, value: list[Hotkey | ColdkeyRef | str] | None):
        if value is not None:
            for i, item in enumerate(value):
                if isinstance(item, (ColdkeyRef, Hotkey)):
                    continue
                match item:
                    case 'all': continue
                    case _:
                        name, *hotkey = str.split(item, '/')
                        match len(hotkey):
                            case 0:
                                value[i] = ColdkeyRef(name=name)
                            case 1:
                                value[i] = Hotkey.model_validate({
                                    'name': name,
                                    'hotkey': hotkey[0]
                                })
                            case _:
                                raise ValueError("invalid format")
        return value

    @pydantic.field_validator('wallet_path', mode='after')
    def validate_wallet_path(cls, value: str) -> str:
        #cls.validate_directory(value, False) # type: ignore
        return value

    @pydantic.model_validator(mode='before')
    def validate_network(cls, values: dict[str, Any]) -> dict[str, Any]:
        chain_endpoint = values.get('chain_endpoint')
        network = values.get('network')
        match network:
            case 'finney': chain_endpoint = FINNEY_ENTRYPOINT
            case 'test': chain_endpoint = FINNEY_TEST_ENTRYPOINT
            case None:
                chain_endpoint = chain_endpoint or FINNEY_ENTRYPOINT
                if chain_endpoint == FINNEY_ENTRYPOINT:
                    network = 'finney'
            case _:
                if not chain_endpoint:
                    raise ValueError(
                        "Provide either BT_SUBNET with a known named network "
                        "or set the BT_CHAIN_ENDPOINT variable."
                    )
                warnings.warn(
                    f"BT_SUBNET specifies unknown network '{network}', "
                    f"but is overridden by BT_CHAIN_ENDPOINT: {chain_endpoint}",
                    UserWarning
                )
        values.update({
            'chain_endpoint': chain_endpoint,
            'network': network
        })
        return values

    @pydantic.model_validator(mode='after')
    def postprocess(self):
        hotkeys: list[Hotkey | ColdkeyRef | str] = list(self.enabled_hotkeys)
        if len(self.enabled_hotkeys) == 1 and hotkeys[0] == 'all': # type: ignore
            hotkeys = []
            for path in self.wallet_path.glob('*/hotkeys/*'):
                m = re.match(r'.*/([\d\w]+)/hotkeys/([\d\w]+)$', str(path))
                if m is None:
                    sys.stderr.write(f"Unsupported hotkey filepath: {path}")
                    sys.stderr.flush()
                    raise SystemExit(1)
                name, hotkey = m.groups()
                hotkeys.append(
                    Hotkey.model_validate({
                        'name': name,
                        'hotkey': hotkey
                    })
                )

        coldkeyrefs = [ref for ref in hotkeys if isinstance(ref, ColdkeyRef)]
        for ref in coldkeyrefs:
            hotkeys.pop(hotkeys.index(ref))
            hotkeys.extend(ref.hotkeys(self.wallet_path, refs_only=True))

        self.enabled_hotkeys = ColonSeparatedList(  # type: ignore
            sorted(hotkeys, key=lambda ref: ref.qualname) # type: ignore
        )
        assert all([isinstance(hotkey, Hotkey) for hotkey in hotkeys])
        return self


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pathlib

import pydantic

from ._hotkey import Hotkey


class ColdkeyRef(pydantic.BaseModel):
    name: str = pydantic.Field(
        default=...
    )

    def hotkeys(
        self,
        wallet_path: pathlib.Path,
        refs_only: bool = True
    ):
        if not refs_only:
            raise NotImplementedError
        hotkeys: list[Hotkey] = []
        for p in sorted(wallet_path.joinpath(self.name).glob('hotkeys/*')):
            hotkeys.append(
                Hotkey.model_validate({
                    'name': self.name,
                    'hotkey': p.stem
                })
            )
        return hotkeys

    def __hash__(self):
        return hash(self.name)


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import inspect
import ipaddress
import pathlib
import os
from typing import Any
from typing import AsyncGenerator
from typing import Awaitable
from typing import Callable
from typing import TypeAlias
from typing import TypeVar
from typing import ParamSpec
from typing import Union

import bittensor
from bittensor.core.settings import DEFAULT_NETWORK
from libcanonical.runtime import MainProcess
from libcanonical.types import FatalException
from libcanonical.utils import deephash

from canonical.ext.bittensor.bases import BaseSoma
from canonical.ext.bittensor.models import Hotkey
from canonical.ext.bittensor.models import SynapseEnvelope
from ..axon import Axon
from ..axon import SynapseResponse
from .._environment import BittensorEnvironment
from .._metagraph import Metagraph
from ._axontransport import AxonTransport
from ._monitor import Monitor
from ._taskrunner import TaskRunner


F = TypeVar('F')
P = ParamSpec('P')
S = TypeVar('S', bound=bittensor.Synapse)

Self = TypeVar('Self', bound='Soma[Any]')

SynapseHandlerType: TypeAlias = Union[
    Callable[..., AsyncGenerator[None | Exception | SynapseResponse, bittensor.Synapse | None]],
    Callable[..., Awaitable[None | SynapseResponse]]
]

SynapseTask: TypeAlias = AsyncGenerator[None | SynapseResponse, bittensor.Synapse | None]


class Soma(MainProcess, BaseSoma[S]):
    """The base class for the :class:`~canonical.ext.bittensor.Validator`
    and :class:`canonical.ext.bittensor.Miner` classes.
    """
    axon_class: type[Axon] = Axon
    config_class: type[BittensorEnvironment] = BittensorEnvironment
    futures: set[asyncio.Future[Any]]
    interval = 0.25
    monitor_class: type[Monitor] = Monitor
    synapse_classes: list[type[bittensor.Synapse]]
    synapse_handlers: dict[str, SynapseHandlerType]
    task_identifier_exclude: set[str] = {'axon', 'dendrite'}

    @classmethod
    def fromenv(cls, name: str, **kwargs: Any):
        env = cls.config_class.model_validate_env({
            **os.environ,
            **{k: v for k, v in kwargs.items() if v is not None}
        })
        if env.netuid is None:
            raise FatalException(
                "Unable to infer netuid from the environment. Set "
                "the BT_SUBNET_UID environment variable to specify "
                "the subnet."
            )
        return cls(
            name=name,
            network=env.network,
            netuid=env.netuid,
            axon_ip=env.axon_ip,
            axon_port=env.axon_port,
            disable_axon=env.axon_disabled,
            hotkeys=list(env.enabled_hotkeys)
        )

    @classmethod
    def register(cls, synapse_class: Any) -> Any:
        def decorator(func: Any):
            func.synapse_class = synapse_class
            return func

        return decorator

    def __init__(
        self,
        name: str,
        netuid: int,
        hotkeys: list[Hotkey] | None = None,
        network: str = DEFAULT_NETWORK,
        axon_ip: ipaddress.IPv4Address = ipaddress.IPv4Address('0.0.0.0'),
        axon_port: int = 8091,
        disable_axon: bool = False,
        wallet_path: pathlib.Path = pathlib.Path('~/.bittensor/wallets'),
        monitoring_bind: str = '0.0.0.0',
        monitoring_port: int = 8888
    ):
        super().__init__(name=name)
        self.axon_ip = axon_ip
        self.axon_port = axon_port
        self.disable_axon = disable_axon
        self.futures = set()
        self.hotkeys = hotkeys or []
        self.metagraph = Metagraph(
            netuid=netuid,
            network=network
        )
        self.monitoring_bind = monitoring_bind
        self.monitoring_port = monitoring_port
        self.netuid = netuid
        self.network = network
        self.synapse_classes = []
        self.synapse_handlers = {}
        self.wallet_path = wallet_path.expanduser()

    def generate_task_id(self, synapse: S) -> str:
        """Generates a task identifier for an incoming synapse. Task identifiers
        are used to identify similar tasks so that the scheduler can optimize
        execution.
        """
        if not synapse.dendrite:
            raise TypeError("The synapse.dendrite attribute can not be None.")
        return deephash(
            (
                synapse.dendrite.hotkey,
                synapse.model_dump(exclude=self.task_identifier_exclude)
             ),
            encode='hex',
            using='sha256'
        )

    def log_status(self):
        if not self.metagraph.wait(timeout=0.1):
            self.logger.critical("Metagraph is not up-to-date (age: %.02fs)", self.metagraph.age)
        self.logger.info(
            'Running on subnet %s (metagraph-age: %.02fs, step: %s, runtime: %.02f)',
            self.netuid,
            self.metagraph.age,
            self.step,
            self.runtime,
        )
        for hotkey in self.hotkeys:
            neuron = self.metagraph.neuron(hotkey.ss58_address)
            match neuron.is_registered():
                case True:
                    self.logger.info(
                        "Hotkey %s: UID %s (incentive: %s, trust: %s, emission: %s)",
                        neuron.hotkey,
                        neuron.uid,
                        neuron.incentive,
                        neuron.trust,
                        neuron.emission
                    )
                case False:
                    self.logger.info(
                        "Hotkey %s: unregistered (incentive: %s, trust: %s, emission: %s)",
                        neuron.hotkey,
                        neuron.incentive,
                        neuron.trust,
                        neuron.emission
                    )

    def must_report(self):
        return all([
            self.runtime > 1,
            self.interval >= 1.0,
            int(self.runtime) % 15 == 0
        ])

    def submit(
        self,
        envelope: SynapseEnvelope[Any],
        future: asyncio.Future[SynapseResponse]
    ) -> None:
        self.tasks.schedule(envelope, future)

    async def configure(self, reloading: bool = False):
        if not reloading:
            self.logger.info("Booting %s", self.name)
            self.monitor = self.monitor_class(
                soma=self,
                host=self.monitoring_bind,
                port=self.monitoring_port
            )
            self.monitor.start()

            for key in self.hotkeys:
                key.load(self.wallet_path, mode='public')
                self.logger.info(
                    "Loaded hotkey %s/%s (ss58: %s)",
                    key.name,
                    key.hotkey,
                    key.ss58_address
                )

            # Ensure that all components have access to the
            # handlers.
            self._register_handlers()

            # Start the metagraph and ensure that the main event loop
            # starts with an up-to-date set of neurons.
            self.metagraph.start()
            self.logger.info(
                "Starting metagraph and awaiting initial update."
            )
            self.metagraph.wait()
            self.logger.info(
                "Commencing main event loop after successful metagraph update."
            )

            # Task runner needs to be available before the transport
            # starts.
            self.tasks = TaskRunner(
                soma=self,
                loop=self.loop,
                handlers=self.synapse_handlers
            )

            # Start transport to receive synapses. For now this is an
            # HTTP server, but other protocols might be considered.
            if self.disable_axon:
                self.transport = AxonTransport(
                    soma=self,
                    synapse_classes=self.synapse_classes
                )
            else:
                self.transport = self.axon_class(
                    soma=self,
                    metagraph=self.metagraph,
                    host=str(self.axon_ip),
                    port=self.axon_port,
                    hotkeys=self.hotkeys,
                    synapse_types=self.synapse_classes
                )
            self.transport.start()
            self.transport.wait()
            self.logger.info(
                "Bound axon transport to %s:%s",
                self.transport.host,
                self.transport.port
            )

            self.logger.info("%s succesfully bootstrapped.", type(self).__name__)

    def add_future(
        self,
        awaitable: asyncio.Future[Any] | Awaitable[Any]
    ):
        if not isinstance(awaitable, asyncio.Future):
            awaitable = asyncio.ensure_future(awaitable)
        self.futures.add(awaitable)

    async def main_event(self) -> None:
        if not self.transport.is_running():
            self.logger.critical("Transport exited unexpectedly.")
            self.stop()
            return
        if self.must_report():
            self.log_status()
        if self.futures:
            _, self.futures = await asyncio.wait(self.futures, timeout=0.05)
        self.add_future(self.tasks.run_all())

    async def teardown(self) -> None:
        self.transport.stop()
        if self.futures:
            _, running = await asyncio.wait(
                self.futures,
                timeout=self.teardown_deadline
            )
            if running:
                self.logger.critical(
                    "%s tasks did not complete within teardown deadline.",
                    len(running)
                )
                for future in running:
                    future.cancel()
        self.metagraph.stop()
        self.monitor.stop()

    async def run_axon(self) -> None:
        raise NotImplementedError

    def _register_handlers(self):
        for _, value in inspect.getmembers(self):
            if not hasattr(value, 'synapse_class'):
                continue
            cls: type[bittensor.Synapse] = getattr(value, 'synapse_class')
            qualname = cls.__name__
            self.synapse_classes.append(cls)
            self.synapse_handlers[qualname] = value
            self.logger.info(
                "Initialized %s handler function.",
                qualname
            )


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
from libcanonical.types import HexEncoded
from starlette.datastructures import Headers

from canonical.ext.bittensor.types import SS58Address


class SynapseHeader(pydantic.BaseModel):
    accept: str | None = pydantic.Field(
        default=None,
        alias='accept'
    )

    content_type: str = pydantic.Field(
        default=...,
        alias='content-type'
    )

    user_agent: str | None = pydantic.Field(
        default=None,
        alias='user-agent'
    )

    accept_encoding: str | None = pydantic.Field(
        default=None,
        alias='accept-encoding'
    )

    version: str = pydantic.Field(
        default='tensorshield.ai/v1',
        alias='bt_version'
    )

    dendrite_hotkey: SS58Address = pydantic.Field(
        default=...,
        alias='bt_header_dendrite_hotkey'
    )

    dendrite_ip: str = pydantic.Field(
        default=...,
        alias='bt_header_dendrite_ip'
    )

    dendrite_nonce: int = pydantic.Field(
        default=...,
        alias='bt_header_dendrite_nonce'
    )

    dendrite_signature: HexEncoded = pydantic.Field(
        default=...,
        alias='bt_header_dendrite_signature'
    )

    dendrite_uuid: str = pydantic.Field(
        default=...,
        alias='bt_header_dendrite_uuid'
    )

    dendrite_version: str = pydantic.Field(
        default=...,
        alias='bt_header_dendrite_version'
    )

    axon_hotkey: SS58Address = pydantic.Field(
        default=...,
        alias='bt_header_axon_hotkey'
    )

    axon_ip: str = pydantic.Field(
        default=...,
        alias='bt_header_axon_ip'
    )

    axon_port: int = pydantic.Field(
        default=...,
        alias='bt_header_axon_port'
    )

    forwared_for: str | None = pydantic.Field(
        default=None,
        alias='x-forwarded-for'
    )

    forwared_protocol: str | None = pydantic.Field(
        default=None,
        alias='x-forwarded-proto'
    )

    computed_body_hash: str = pydantic.Field(
        default=...,
        alias='computed_body_hash'
    )

    subnet_uid: int | None = pydantic.Field(
        default=None,
        alias='bittensor-subnet-uid'
    )

    timeout: float = pydantic.Field(
        default=...,
        alias='timeout'
    )

    @classmethod
    def model_validate_headers(cls, headers: Headers):
        return cls.model_validate(headers)

    def model_dump_headers(self):
        headers = self.model_dump(
            by_alias=True,
            exclude_none=True
        )
        return {k: str(v) for k, v in headers.items()}


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
from cryptography.hazmat.primitives.asymmetric import ed25519
from libcanonical.types import HexEncoded

from canonical.ext.bittensor.types import SS58Address
from ._hotkeyref import HotkeyRef


class HotkeyPublicKey(HotkeyRef):
    public_key: HexEncoded = pydantic.Field(
        default=...,
        alias='publicKey'
    )

    ss58_address: SS58Address = pydantic.Field(
        default=...,
        alias='ss58Address'
    )

    @property
    def public(self):
        return ed25519.Ed25519PublicKey.from_public_bytes(self.public_key)


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import socket

import pydantic
from bittensor.core.async_subtensor import AsyncSubtensor
from libcanonical.bases import PollingExternalState
from libcanonical.utils import retry

from .models import Neuron
from .models import NeuronList
from ._const import DEFAULT_NETWORK



NETWORK_ERRORS: list[type[BaseException]] = [
    ConnectionError,
    TimeoutError,
    socket.gaierror
]


class Metagraph(PollingExternalState):
    """Keeps a real-time state of the Bittensor metagraph on the network
    and subnet specified.
    """
    __module__: str = 'canonical.ext.bittensor'
    block: int = 0
    neurons: NeuronList
    subtensor: AsyncSubtensor | None = None

    def __init__(
        self,
        network: str = DEFAULT_NETWORK,
        netuid: int = 0,
        interval: float = 10,
        immediate: bool = False,
    ):
        self.network = network
        self.netuid = netuid
        self.neurons = NeuronList(netuid=self.netuid, block=0)
        super().__init__(interval=interval, immediate=immediate)

    def get_subtensor(self) -> AsyncSubtensor:
        return AsyncSubtensor(
            network=self.network
        )

    def neuron(
        self,
        hotkey: str,
        validator_permit: bool | None = True
    ) -> Neuron:
        adapter: pydantic.TypeAdapter[Neuron] = pydantic.TypeAdapter(Neuron)
        neuron = self.neurons.get(str(hotkey))
        if neuron is None:
            neuron = adapter.validate_python({
                'coldkey': '',
                'hotkey': hotkey,
                'uid': -1,
                'validator_permit': validator_permit
            })
        return neuron

    def setready(self):
        if len(self.neurons.items) > 0:
            super().setready()

    def stop(self):
        self.logger.info("Stopping metagraph.")
        self.must_stop = True
        self.join()

    @retry(NETWORK_ERRORS, reason='Bittensor network unreachable.')
    async def main_event(self) -> None:
        assert self.subtensor is not None
        last_block = self.block
        self.block = await self.subtensor.get_current_block()
        if last_block == self.block:
            return
        try:
            changed, joined, dropped = await self.neurons.update(
                subtensor=self.subtensor,
                netuid=self.netuid,
                block=self.block
            )
            self.log(
                'INFO', "Metagraph update block %s (joined: %s, changed: %s, dropped: %s)",
                self.block,
                len(joined),
                len(changed),
                len(dropped)
            )
        except Exception as e:
            self.logger.exception(
                "Caught fatal %s: %s",
                type(e).__name__,
                repr(e)
            )

    @retry(NETWORK_ERRORS, reason='Bittensor network unreachable.')
    async def setup(self, reloading: bool = False) -> None:
        if reloading:
            if self.subtensor is not None:
                await self.subtensor.close()
        self.subtensor = self.get_subtensor()
        await self.subtensor.__aenter__()

    @retry(NETWORK_ERRORS, max_attempts=5, delay=1, reason='Bittensor network unreachable.')
    async def teardown(
        self,
        exception: BaseException | None = None
    ) -> bool:
        if self.subtensor is not None:
            await self.subtensor.close()
        return False


# Copyright (C) 2022-2025 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pathlib
from typing import Literal

import pydantic

from ._hotkey import Hotkey


class HotkeySet(pydantic.BaseModel):
    model_config = {'populate_by_name': True}

    items: dict[str, Hotkey] = pydantic.Field(
        default_factory=dict
    )

    def load(
        self,
        path: pathlib.Path,
        mode: Literal['public', 'private', 'phrase']
    ):
        for hotkey in self.items.values():
            hotkey.load(path, mode=mode)